
//////////////////////////////////////////////////////////////////////
// Command-line wrapper for RegX regular-expression module
// 2005/3/3 -- Copyright (c) 2003 Peter J. Goodeve -- All Rights Reserved
//
//	Command line:
//		scan [options] pattern [file...]
//
//	-- if no file, reads from standard input. "-" as a filename
//	also specifies stdin.
//
//	Options:
//			-s	Find *Shortest* match -- default is longest.
//			-p	Show Position of match (bounding character-numbers
//				of the imput stream -- first char to first non-match).
//			-t	Show Text of match (default unless -p is specified)
//			-l	Respect Lines -- don't put each matching text on its
//				own line (newlines only if actually in the match).
//			-a	All characters -- *including* newline -- match '.'
//				-- default is that newline does not match.
//			-n	'$' *only* matches at an actual newline.  By default
//				end-of text is always taken to be an end-of-line.
//				(Of course it matches the (null) EOL itself, not
//				the newline, which does not appear in the match.)
//			-i	Ignore case.
//			-v	Inverse match -- output unmatched segments; only output
//				matched sections via template.
//
//////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "regx.h"
#include "filestrip.h"
#include "utfstrip.h"


// #define RXDEBUG

#ifdef	RXDEBUG
 #define DPRINTF(s) printf s
 char * iname(int t);
#else
 #define DPRINTF(s)
#endif

// for temporary quick-checks...
// Always on:
#define XDPRINTF(s) printf s
// Always off:
#define ZDPRINTF(s)

// Keep this in step with resource!
#define VERSION "matt 1.4 (2006/1/23)"

#define SEGMAX 20

struct TmpltFmt {
	char *tseg[SEGMAX];
	int tlen[SEGMAX];
	int subid[SEGMAX];
	int fmtlen;
	TmpltFmt(char *fmt);
	void print(FileStrip &fs, Rangeset &sel, char *fname);
};


bool printSeg(char *seg, int len) {
	fwrite(seg, len, 1, stdout);
	fflush(stdout);
	return true;
}

long	matches = 0,	// keep a (total) count
		filemcnt;			// reset for each file

const char* errstr[] = {
	"Regular Expression too long",	//	EXPR_TOO_LONG
	"Badly placed left parenthesis",	//	LEFT_PAR_ERR
	"Badly placed right parenthesis",	//	RIGHT_PAR_ERR
	"Missing Operator",	//	MISSING_OP
	"Badly formed regular expression",	//	BAD_REG_EXPR
	"Internal Error!",	//	BAD_CODE
	"Overflow while scanning",	//	OVERFLOW
	"Scan failed!"	//	SCAN_FAIL
};

int main(int argc, char **argv)
{
	bool	shortest = false,	// grab longest by default (not normally across lines)
			printposns = false,
			printtext = false,	// but will be set true by default
			printnames = false,	// output each file name
			countonly = false,	// output only a count of the matches
			forcenl = true,		// by default each match is on a new line
			nlisany = false,	// if true, '.' matches newline
			eotiseol = true,	// by default end-of-text matched by '$'
			nocase = false,
			useutf = true,
			unmatched = false;	// if set, unmatched segments will be output
			
	TmpltFmt *outformat = NULL;		// output template placed here
	char * patfile = NULL;			// if pattern is obtained from file
	char tmpltline[200]="\0";	// template obtained from patfile
	RegX rx;
	do {
		int opt = getopt(argc, argv, "sptnailcvz8Vo:f:");
		if (opt == EOF) break;
		switch(opt) {
		case 's':
			shortest = true;
			break;
		case 'p':
			printposns = true;
			break;
		case 't':
			printtext = true;
			break;
		case 'n':
			forcenl = false;
			break;
		case 'a':
			nlisany = true;
			break;
		case 'z':
			eotiseol = false;
			break;
		case 'i':
			nocase = true;
			break;
		case '8':
			useutf = false;
			break;
		case 'l':
			printnames = true;
			break;
		case 'c':
			countonly = true;
			break;
		case 'v':
			unmatched = true;
			break;
		case 'o':
			outformat = new TmpltFmt(optarg);
			break;
		case 'f':
			patfile = optarg;
			break;
		case 'V':
			printf("%s\n", VERSION);
			exit(0);
		}
	} while (true);
	if (!(printposns || printnames ||countonly)) printtext = true;
	if (outformat) printposns = printtext = false;
		
	if (optind >= argc) {
		fprintf(stderr, "Format:\n   matt [options] pattern [-o template] files...\nOptions:\n");
		fprintf(stderr, "-s	 find shortest matching string (default longest)\n");
		fprintf(stderr, "-p	 print begin/end char positions of match in text\n");
		fprintf(stderr, "-8	 Scan as (8-bit) bytes rather than default UTF\n");
		fprintf(stderr, "-t	 force match-text printing (with e.g. '-p')\n");
		fprintf(stderr, "-n	  no forced newline after each match (template always verbatim)\n");
		fprintf(stderr, "-a	 '.' matches newline as well (normally matches all but...)\n");
		fprintf(stderr, "-z	 assume end-of-line at EOF\n");
		fprintf(stderr, "-i	 case insensitive\n");
		fprintf(stderr, "-l	 list just the file name of any file with a match\n");
		fprintf(stderr, "-v	 inverse match -- output unmatched segments\n");
		fprintf(stderr, "-c	 output only a count of all matches in the file\n");
		fprintf(stderr, "-V	 print out version and exit\n");
		fprintf(stderr, "-f file	 get pattern (and optionally template) from file\n");
		fprintf(stderr, "\nTemplate Specials:\n");
		fprintf(stderr, "$0	entire matched string\n");
		fprintf(stderr, "$1-$9	matched subsegment n\n");
		fprintf(stderr, "$b	position (chars) of beginning of match\n");
		fprintf(stderr, "$e	position (chars) after end of match\n");
		fprintf(stderr, "$f	current filename\n");
		fprintf(stderr, "$n	number of the match in the current file\n");
		fprintf(stderr, "$t	total so far of all matches in all files scanned\n");
		fprintf(stderr, "$(n...)	if subpattern n matched [n=1..9], insert string\n");
		fprintf(stderr, "$(-n...)	if subpattern n [n=1..9] not in match, insert string\n");
		fprintf(stderr, "$(+n...)	insert subelement n if found, otherwise insert string\n");
		return 0;
	}

	rx.shortest = shortest;
	rx.nlisany = nlisany;
	rx.eotiseol = eotiseol;
	rx.nocase = nocase;

	 try {
		if (patfile) {
			FILE *pf = fopen(patfile, "r");
			char patline[200];
			if (!pf) {
				fprintf(stderr, "Couldn't open pattern file %s!\n", patfile);
				return -1;	// find better error number later!
			}
			fgets(patline, 200, pf);
			if (*patline == '#') fgets(patline, 200, pf);	// first line may be comment
			fgets(tmpltline, 200, pf);	// if it is present
			fclose (pf);
			DPRINTF(("compiling pat %s from file %s\n", patline, patfile);)
			char *lastpos = patline+strlen(patline)-1;
			if (*lastpos == '\n') *lastpos = '\0';
			UtfStrip pat(patline, useutf);
			rx.compile(&pat);
			if (*tmpltline && !outformat) {	// always inhibited by command-line arg
				lastpos = tmpltline+strlen(tmpltline)-1;
				if (*lastpos == '\n') *lastpos = '\0';
				if (*tmpltline) outformat = new TmpltFmt(tmpltline);	// ignore blank line
			}
		} else {
			DPRINTF(("compiling pat arg %s\n", argv[optind++]);)
			UtfStrip pat(argv[optind++], useutf);
			rx.compile(&pat);
		}
		DPRINTF(("============compiled pat============\n");)
	 }
	 catch(RegX::Error &e) {
		if (e.c)
			fprintf(stderr, "Regexp Compile Error %d at '%c' -- %s\n",
			 e.code, e.c, errstr[e.code]);
		else if (e.code > RegX::BAD_CODE)
			fprintf(stderr, "Scan Error %d  -- %s\n",
			 e.code, errstr[e.code]);
		else
			fprintf(stderr, "Regexp Compile Error %d -- %s\n",
			 e.code, errstr[e.code]);
		return 2;
	 }
	DPRINTF(("searching %d files\n", argc - optind);)

	if (argc > optind + 1) printnames = true;
	ZDPRINTF(("printnames %d outformat = %p\n", printnames, outformat);)
	FILE *f = NULL;
	if (optind == argc) {
		f = stdin;
		optind--;	// hack k...k
		DPRINTF(("using default stdin -- argc=%d optind=%d\n", argc, optind);)
	}
	while (optind < argc) {
		bool nameshown = false;
		filemcnt = 0;	// reset match count
		if (!f) {
			ZDPRINTF(("opening %s\n", argv[optind]);)
			if (!strcmp(argv[optind], "-")) f = stdin;
			else f = fopen(argv[optind], "r");
		}
		if (!f) {
			fprintf(stderr, "couldn't open %s\n", argv[optind++]);
			continue;
		}
	 try {
		FileStrip tests(f, useutf);
		tests.segProc = printSeg;
		tests.procDiscard = unmatched;	// true on '-v'
		long res=0, matchstart=0, matchend=0, len=0;
		long keep=0;
		do {
			DPRINTF(("scanning from %d\n", tests.pos());)
			tests.keepAfter(keep-1);	// We have to be able to look back one char!
			res = rx.scan(&tests, tests.pos());
			DPRINTF(("return from scan res=%d  pos %d\n", res, tests.pos());)
			matchstart = rx.sel.p[0].p1;
			matchend = rx.sel.p[0].p2;
			keep = tests.keepAfter()+1;	// may have been changed under us
			DPRINTF(("Unmatched from %d to %d\n", keep, matchstart);)
			if (unmatched) {
				if (matchstart < 0) matchstart = tests.pos();
				DPRINTF(("outputting unmatched from %d to %d\n", keep, matchstart);)
				tests.processSeg(keep, matchstart);
			}
			keep = matchend;
			if (res) {
				len = matchend - matchstart;
				if (printnames && !nameshown) {
					printf("%s%s\n", argv[optind],
						printtext || printposns? ":" : "");
					nameshown = true;
				}
				if (printposns)
					printf("%d-%d%s", matchstart, matchend,
						(printtext && len)? ": " : "\n");
				if (len == 0) {
						matchend++;	// at least bump past...
				}
				else {
					matches++;
					filemcnt++;
					if (outformat) {
						outformat->print(tests, rx.sel, argv[optind]);
					}
					else if (printtext && !unmatched) {
						tests.processSeg(matchstart, matchend);
						if (forcenl) fputs("\n", stdout);
					}
				}
			}
			// restart from where we left off:
		} while (res && tests.pos(matchend));
		if (countonly && filemcnt)
				printf("%d\n", filemcnt);
	 }
	 catch(RegX::Error &e) {
		if (e.c)
			fprintf(stderr, "regexp scan error %d '%c'\n", e.code, e.c);
		else
			fprintf(stderr, "regexp scan error %d\n", e.code);
		return 3;
	 }
	 f = NULL;	// so we open another...
	 optind++;
	}

	delete outformat;	// if used
	return matches? 0 : 1;
}

static char procEsc(char *&cp) {
	char c = *cp;
	if (c == '\\') {
		c= *++cp;	// actual char to decode
		switch (c) {
			case 'n': c='\n'; break;	// newline
			case 't': c='\t'; break;	// tab
			case 'r': c='\r'; break;	// carriage-return
			case 'b': c='\b'; break;	// backspace
			case 'f': c='\f'; break;	// formfeed
			case 'v': c='\v'; break;	// vertical tab
			case 'a': c='\a'; break;	// 'bell' (alert) char
			case 'q': c='\''; break;	// single-quote
			case '0':
			case '1':
			case '2':
			case '3':
			case '4':
			case '5':
			case '6':
			case '7':
					{
						DPRINTF(("first octal entry char %c\n", c);)
						c = 0;
						for (int i=3 ; i; i--) {
							int tc = *cp;
							DPRINTF(("octal entry char %c\n", tc);)
							if (tc < '0' || tc > '7') break;
							c = (c << 3) + tc - '0';
							cp++;
						}
						DPRINTF(("value=%x next char %c\n", c, *cp);)
						cp--;	// back up because loop will increment!
					}
						; break;
			case 'x':
			case 'X':
					{
						c = 0;
						DPRINTF(("fhex spec char %c\n", c);)
						cp++;
						for (int i=2 ; i; i--) {	// only one byte for output!
							int tc = *cp;
							DPRINTF(("hex entry char %c\n", tc);)
							if (tc >= '0' && tc <= '9') tc -= '0';
							else if (tc >= 'A' && tc <= 'F') tc += 0xA - 'A';
							else if (tc >= 'a' && tc <= 'f') tc += 0xA - 'a';
							else break;
							c = (c << 4) + tc;
							cp++;
						}
						DPRINTF(("value=%x next char %c\n", c, *cp);)
						cp--;	// back up because loop will increment!
					}
						; break;
		}	// other chars (esp '$') escape to themselves...
	}
	return c;
}

enum {
	SUB_MISSING=0x100,
	SUB_ALT=0x200
};

TmpltFmt::TmpltFmt(char *fmt) {	// caution -- fmt string may get changed...
	DPRINTF(("building format %s at %p\n", fmt, fmt);)
	memset(this, 0, sizeof(TmpltFmt));
	int segno = 0;
	tseg[0] = fmt;
	char *fp, *tp;

	for (fp = tp = fmt; *fp && segno < SEGMAX; fp++)
		if (*fp == '$') {
			DPRINTF(("format special -- segment %d len %d sub %c at %p\n",
					segno, tlen[segno], *fp, fp);)
			if (tlen[segno]) segno++;	//don't waste slots
			tp = fp;		// resync
			tseg[segno] = NULL;
			if (*++fp == '(') {
				char id = *++fp;
				int mode = 0;
				if (id == '-') {
					id = *++fp;
					mode = SUB_MISSING;
				}
				else if (id == '+') {
					id = *++fp;
					mode = SUB_ALT;
				}
				if (id < '1' || id > '9') break;	// must be valid digit
				subid[segno] = id | mode;
				if (*++fp) tseg[segno] = fp;	// protect against bad format
				for (tp=fp; *fp && *fp != ')'; tp++, fp++) { 
					*tp = procEsc(fp);
					tlen[segno]++;
				}
			}
			else subid[segno] = *fp;		// ID is now just the char
			tseg[++segno] = tp = fp+1;	// start of next segment
		}
		else {
			DPRINTF(("text segment char %c from %p to %p\n", *fp, fp, tp);)
			tlen[segno]++;
			*tp++ = procEsc(fp);
		}
	DPRINTF(("format done -- last segment %d len %d last fp %p tp %p\n",
					segno, tlen[segno], fp, tp);)
}


void TmpltFmt::print(FileStrip &fs, Rangeset &sel, char *fname)
{
	for (int i=0; (tseg[i] || subid[i])  && i < SEGMAX; i++) {
		if (tseg[i] && *tseg[i] && !subid[i]) {
			DPRINTF(("printing text from %p length %d\n",
					 tseg[i], tlen[i]);)
			printSeg(tseg[i], tlen[i]);
		}
		else if (subid[i]) {
			int n = subid[i] & 0xff;
			int mode = subid[i] & 0x300;
			switch (n) {
				case '0': case '1': case '2': case '3': case '4':
				case '5': case '6': case '7': case '8': case '9':
					n -= '0';
	     /* Test logic: [someday I'll explain this... (:-/)]
	         tlen    0   -   +   sel    act
	         ----   ---  -----   ---    ---
	          no                 yes  print sel      A
	          no                 no   skip           B
	          yes    y           yes  print text     C
	          yes    y           no   skip           D
	          yes        y       yes  skip           E
	          yes        y       no   print text     F
	          yes            y   yes  print sel      G
	          yes            y   no   print text     H
	        Test order:
	          no                 yes  print sel      A
	          yes            y   yes  print sel      G
	          yes    y           yes  print text     C
	          yes        y       no   print text     F
	          yes            y   no   print text     H
	          yes    y           no   skip           D
	          yes        y       yes  skip           E
	          no                 no   skip           B
	     */
					if (sel.p[n].p2) {
						if (!tlen[i] || mode == 0x200)	// A,G
							fs.processSeg(sel.p[n].p1, sel.p[n].p2);
						else if (!mode)					// C
							 printSeg(tseg[i], tlen[i]);
					}
					else if (tlen[i] && mode) {	// F,H
							 printSeg(tseg[i], tlen[i]);
					}
					// B,D,E
						break;
				case 'f':
						printSeg(fname, strlen(fname));
						break;
				case 'b':
						fprintf(stdout, "%d", sel.p[0].p1);
						break;
				case 'e':
						fprintf(stdout, "%d", sel.p[0].p2);
						break;
				case 'n':
						fprintf(stdout, "%d", filemcnt);
						break;
				case 't':
						fprintf(stdout, "%d", matches);
						break;
				default:
						fputc(n, stdout);
						break;
			}
		}
	}
}

